import subprocess
import sys
from typing import List, Set

import click

from ape import config
from ape.cli import ape_cli_context, skip_confirmation_option
from ape.plugins import clean_plugin_name, plugin_manager
from ape.utils import get_package_version
from ape_plugins.utils import (
    FIRST_CLASS_PLUGINS,
    SECOND_CLASS_PLUGINS,
    extract_module_and_package_install_names,
    is_plugin_installed,
)


@click.group(short_help="Manage ape plugins")
def cli():
    """
    Command-line helper for managing installed plugins.
    """


def _display_section(header: str, lines: List[Set[str]]):
    click.echo(header)
    for output in lines:
        if output:
            formatted_output = _format_output(output)
            click.echo("  {}".format("\n  ".join(formatted_output)))


def _format_output(plugins_list: Set[str]) -> Set:
    output = set()
    for i in plugins_list:
        text = i.replace("ape_", "")
        output.add(text)
    return output


@cli.command(name="list", short_help="Display plugins")
@click.option(
    "-a",
    "--all",
    "display_all",
    default=False,
    is_flag=True,
    help="Display all plugins installed and available (including Core)",
)
@ape_cli_context()
def _list(cli_ctx, display_all):
    installed_first_class_plugins = set()
    installed_second_class_plugins = set()
    installed_second_class_plugins_no_version = set()
    installed_third_class_plugins = set()

    for name, plugin in plugin_manager.list_name_plugin():
        version = get_package_version(name)

        if name in FIRST_CLASS_PLUGINS:
            if not display_all:

                continue  # NOTE: Skip 1st class plugins unless specified
            installed_first_class_plugins.add(f"{name}")

        elif name in SECOND_CLASS_PLUGINS:
            installed_second_class_plugins.add(f"{name}     {version}")
            installed_second_class_plugins_no_version.add(f"{name}")

        elif name not in FIRST_CLASS_PLUGINS or name not in SECOND_CLASS_PLUGINS:
            installed_third_class_plugins.add(f"{name}      {version})")
        else:
            cli_ctx.logger.error(f"{name} is not a plugin.")

    sections = {}

    # First Class Plugins
    if display_all:
        sections["Installed Core Plugins"] = [installed_first_class_plugins]

    # Second and Third Class Plugins
    available_second = list(SECOND_CLASS_PLUGINS - installed_second_class_plugins_no_version)

    if installed_second_class_plugins:
        sections["Installed Plugins"] = [
            installed_second_class_plugins,
            installed_third_class_plugins,
        ]

    elif not display_all:
        # user has no plugins installed | cant verify installed plugins
        if available_second:
            click.echo("There are available plugins to install, use -a to list all plugins ")

    if display_all:

        available_second_output = _format_output(available_second)
        if available_second_output:

            sections["Available Plugins"] = [available_second_output]

        else:
            if SECOND_CLASS_PLUGINS:
                click.echo("You have installed all the available Plugins\n")

    for i in range(0, len(sections)):
        header = list(sections.keys())[i]
        output = sections[header]
        _display_section(f"{header}:", output)

        if i < (len(sections) - 1):
            click.echo()


@cli.command(short_help="Install an ape plugin")
@click.argument("plugin")
@click.option("-v", "--version", help="Specify version (Default is latest)")
@skip_confirmation_option(help="Don't ask for confirmation to add the plugin")
@ape_cli_context()
def add(cli_ctx, plugin, version, skip_confirmation):
    if plugin.startswith("ape"):
        cli_ctx.abort(f"Namespace 'ape' in '{plugin}' is not required")

    # NOTE: Add namespace prefix (prevents arbitrary installs)
    plugin = f"ape_{clean_plugin_name(plugin)}"

    if version:
        plugin = f"{plugin}=={version}"

    if plugin in FIRST_CLASS_PLUGINS:
        cli_ctx.abort(f"Cannot add 1st class plugin '{plugin}'")

    elif is_plugin_installed(plugin):
        cli_ctx.abort(f"Plugin '{plugin}' already installed")

    elif (
        plugin in SECOND_CLASS_PLUGINS
        or skip_confirmation
        or click.confirm(f"Install unknown 3rd party plugin '{plugin}'?")
    ):
        cli_ctx.logger.info(f"Installing {plugin}...")
        # NOTE: Be *extremely careful* with this command, as it modifies the user's
        #       installed packages, to potentially catastrophic results
        # NOTE: This is not abstracted into another function *on purpose*
        subprocess.call([sys.executable, "-m", "pip", "install", "--quiet", plugin])


@cli.command(short_help="Install all plugins in the local config file")
@ape_cli_context()
@skip_confirmation_option("Don't ask for confirmation to install the plugins")
def install(cli_ctx, skip_confirmation):
    plugins = config.get_config("plugins") or []
    for plugin in plugins:
        module_name, package_name = extract_module_and_package_install_names(plugin)
        if not is_plugin_installed(module_name) and (
            module_name in SECOND_CLASS_PLUGINS
            or skip_confirmation
            or click.confirm(f"Install unknown 3rd party plugin '{package_name}'?")
        ):
            cli_ctx.logger.info(f"Installing {package_name}...")
            # NOTE: Be *extremely careful* with this command, as it modifies the user's
            #       installed packages, to potentially catastrophic results
            # NOTE: This is not abstracted into another function *on purpose*
            subprocess.call([sys.executable, "-m", "pip", "install", "--quiet", f"{package_name}"])


@cli.command(short_help="Uninstall an ape plugin")
@click.argument("plugin")
@skip_confirmation_option("Don't ask for confirmation to remove the plugin")
@ape_cli_context()
def remove(cli_ctx, plugin, skip_confirmation):
    if plugin.startswith("ape"):
        cli_ctx.abort(f"Namespace 'ape' in '{plugin}' is not required")

    # NOTE: Add namespace prefix (match behavior of ``install``)
    plugin = f"ape_{clean_plugin_name(plugin)}"

    if not is_plugin_installed(plugin):
        cli_ctx.abort(f"Plugin '{plugin}' is not installed")

    elif plugin in FIRST_CLASS_PLUGINS:
        cli_ctx.abort(f"Cannot remove 1st class plugin '{plugin}'")

    elif skip_confirmation or click.confirm(
        f"Remove plugin '{plugin} ({get_package_version(plugin)})'"
    ):
        # NOTE: Be *extremely careful* with this command, as it modifies the user's
        #       installed packages, to potentially catastrophic results
        # NOTE: This is not abstracted into another function *on purpose*
        subprocess.call([sys.executable, "-m", "pip", "uninstall", "--quiet", "-y", plugin])



import pytest

from ape.api import TransactionAPI, TransactionType
from ape.api.accounts import AccountAPI
from ape.exceptions import AccountsError
from ape.types import AddressType

from ..conftest import TEST_ADDRESS


class _TestAccountAPI(AccountAPI):
    can_sign: bool

    @property
    def address(self):
        return AddressType(TEST_ADDRESS)

    def sign_message(self, msg):
        return "test-signature" if self.can_sign else None

    def sign_transaction(self, txn):
        return "test-signature" if self.can_sign else None


@pytest.fixture
def test_account_api_no_sign(mock_account_container_api, mock_provider_api):
    account = _TestAccountAPI(mock_account_container_api, can_sign=False)
    account._provider = mock_provider_api
    return account


@pytest.fixture
def test_account_api_can_sign(mock_account_container_api, mock_provider_api):
    account = _TestAccountAPI(mock_account_container_api, can_sign=True)
    account._provider = mock_provider_api
    return account


class TestAccountAPI:
    def test_txn_nonce_less_than_accounts_raise_tx_error(
        self, mocker, mock_provider_api, test_account_api_can_sign
    ):
        mock_transaction = mocker.MagicMock(spec=TransactionAPI)

        # Differing nonces
        mock_provider_api.get_nonce.return_value = 1
        mock_transaction.nonce = 0

        with pytest.raises(AccountsError) as err:
            test_account_api_can_sign.call(mock_transaction)

        assert str(err.value) == "Invalid nonce, will not publish."

    def test_not_enough_funds_raises_error(
        self, mocker, mock_provider_api, test_account_api_can_sign
    ):
        mock_transaction = mocker.MagicMock(spec=TransactionAPI)
        mock_provider_api.get_nonce.return_value = mock_transaction.nonce = 0
        mock_transaction.type = TransactionType.STATIC
        mock_transaction.gas_price = 0

        # Transaction costs are greater than balance
        mock_transaction.total_transfer_value = 1000000
        mock_provider_api.get_balance.return_value = 0

        with pytest.raises(AccountsError) as err:
            test_account_api_can_sign.call(mock_transaction)

        expected = (
            "Transfer value meets or exceeds account balance.\n"
            "Are you using the correct provider/account combination?\n"
            "(transfer_value=1000000, balance=0)."
        )
        assert str(err.value) == expected

    def test_transaction_not_signed_raises_error(
        self, mocker, mock_provider_api, test_account_api_no_sign
    ):
        mock_transaction = mocker.MagicMock(spec=TransactionAPI)
        mock_provider_api.get_nonce.return_value = mock_transaction.nonce = 0
        mock_transaction.total_transfer_value = mock_provider_api.get_balance.return_value = 1000000
        mock_transaction.type = TransactionType.STATIC
        mock_transaction.gas_price = 0

        with pytest.raises(AccountsError) as err:
            test_account_api_no_sign.call(mock_transaction)

        assert str(err.value) == "The transaction was not signed."

    def test_transaction_when_no_gas_limit_calls_estimate_gas_cost(
        self, mocker, mock_provider_api, test_account_api_can_sign
    ):
        mock_transaction = mocker.MagicMock(spec=TransactionAPI)
        mock_transaction.type = TransactionType.STATIC
        mock_transaction.gas_price = 0
        mock_transaction.gas_limit = None  # Causes estimate_gas_cost to get called
        mock_provider_api.get_nonce.return_value = mock_transaction.nonce = 0
        mock_transaction.total_transfer_value = mock_provider_api.get_balance.return_value = 1000000
        mock_transaction.signature.return_value = "test-signature"
        test_account_api_can_sign.call(mock_transaction)
        mock_provider_api.estimate_gas_cost.assert_called_once_with(mock_transaction)



import os
from typing import Dict, Set, Tuple

from github import Github

from ape.__modules__ import __modules__
from ape.exceptions import ConfigError
from ape.logging import logger

# Plugins maintained OSS by ApeWorX (and trusted)
FIRST_CLASS_PLUGINS = set(__modules__)

SECOND_CLASS_PLUGINS: Set[str] = set()

# TODO: Should the github client be in core?
# TODO: Handle failures with connecting to github (potentially cached on disk?)
if "GITHUB_ACCESS_TOKEN" in os.environ:
    author = Github(os.environ["GITHUB_ACCESS_TOKEN"]).get_organization("ApeWorX")

    SECOND_CLASS_PLUGINS = {
        repo.name.replace("-", "_") for repo in author.get_repos() if repo.name.startswith("ape-")
    }

else:
    logger.warning("$GITHUB_ACCESS_TOKEN not set, unable to list all plugins")


def is_plugin_installed(plugin: str) -> bool:
    try:
        __import__(plugin)
        return True
    except ImportError:
        return False


def extract_module_and_package_install_names(item: Dict) -> Tuple[str, str]:
    """
    Extracts the module name and package name from the configured
    plugin. The package name includes `==<version>` if the version is
    specified in the config.
    """
    try:
        name = item["name"]
        module_name = f"ape_{name.replace('-', '_')}"
        package_install_name = f"ape-{name}"
        version = item.get("version")

        if version:
            package_install_name = f"{package_install_name}=={version}"

        return module_name, package_install_name

    except Exception as err:
        raise _get_config_error() from err


def _get_config_error() -> ConfigError:
    expected_format = "plugins:\n  - name: <plugin-name>\n    version: <plugin-version>  # optional"
    return ConfigError(f"Config item mis-configured. Expected format:\n\n{expected_format}\n")



from typing import Optional

import pytest

import ape
from ape.api import TestProviderAPI
from ape.logging import logger

from .contextmanagers import RevertsContextManager


class PytestApeRunner:
    def __init__(self, config, project, networks):
        self.config = config
        self.project = project
        self.networks = networks
        self._warned_for_missing_features = False
        ape.reverts = RevertsContextManager

    @property
    def _network_choice(self) -> str:
        """
        The option the user providers via --network (or the default).
        """
        return self.config.getoption("network")

    @property
    def _provider(self) -> Optional[TestProviderAPI]:
        """
        The active provider.
        """
        return self.networks.active_provider

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item, nextitem):
        snapshot_id = None

        # Try to snapshot if the provider supported it.
        if hasattr(self._provider, "snapshot"):
            try:
                snapshot_id = self._provider.snapshot()
            except NotImplementedError:
                self._warn_for_unimplemented_snapshot()
                pass
        else:
            self._warn_for_unimplemented_snapshot()

        yield

        # Try to revert to the state before the test began.
        if snapshot_id:
            self._provider.revert(snapshot_id)

    def _warn_for_unimplemented_snapshot(self):
        if self._warned_for_missing_features:
            return

        logger.warning(
            "The connected provider does not support snapshotting. "
            "Tests will not be completely isolated."
        )
        self._warned_for_missing_features = True

    def pytest_sessionstart(self):
        """
        Called after the `Session` object has been created and before performing
        collection and entering the run test loop.

        Removes `PytestAssertRewriteWarning` warnings from the terminalreporter.
        This prevents warnings that "the `ape` library was already imported and
        so related assertions cannot be rewritten". The warning is not relevant
        for end users who are performing tests with ape.
        """
        reporter = self.config.pluginmanager.get_plugin("terminalreporter")
        warnings = reporter.stats.pop("warnings", [])
        warnings = [i for i in warnings if "PytestAssertRewriteWarning" not in i.message]
        if warnings and not self.config.getoption("--disable-warnings"):
            reporter.stats["warnings"] = warnings

    @pytest.hookimpl(trylast=True, hookwrapper=True)
    def pytest_collection_finish(self, session):
        """
        Called after collection has been performed and modified.
        """
        outcome = yield

        # Only start provider if collected tests.
        if not outcome.get_result() and session.items and not self.networks.active_provider:
            self.networks.active_provider = self.networks.get_provider_from_choice(
                self._network_choice
            )
            self.networks.active_provider.connect()

    def pytest_sessionfinish(self):
        """
        Called after whole test run finished, right before returning the exit
        status to the system.
        """
        if self._provider is not None:
            self._provider.disconnect()



from typing import Any, List, Optional, Type

import click
from click import Context, Parameter

from ape import accounts, networks
from ape.api.accounts import AccountAPI
from ape.exceptions import AccountsError


def _get_account_by_type(account_type: Optional[Type[AccountAPI]] = None) -> List[AccountAPI]:
    return list(accounts) if not account_type else accounts.get_accounts_by_type(account_type)


class Alias(click.Choice):
    """Wraps ``click.Choice`` to load account aliases for the active project at runtime.

    Provide an ``account_type`` to limit the type of account to choose from.
    Defaults to all account types in ``choices()``.
    """

    name = "alias"

    # noinspection PyMissingConstructor
    def __init__(self, account_type: Optional[Type[AccountAPI]] = None):
        # NOTE: we purposely skip the constructor of `Choice`
        self.case_sensitive = False
        self._account_type = account_type

    @property
    def choices(self) -> List[str]:  # type: ignore
        options = _get_account_by_type(self._account_type)
        return [a.alias for a in options if a.alias is not None]


class PromptChoice(click.ParamType):
    """
    A choice option or argument from user selection.
    """

    def __init__(self, choices):
        self.choices = choices

    def print_choices(self):
        choices = dict(enumerate(self.choices, 0))
        for choice in choices:
            click.echo(f"{choice}. {choices[choice]}")
        click.echo()

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Optional[Any]:
        # noinspection PyBroadException
        try:
            choice_index = int(value)
            if choice_index < 0:
                self.fail("Invalid choice", param=param)

            return self.choices[choice_index]

        except Exception:
            return self.fail_from_invalid_choice(param)

    def fail_from_invalid_choice(self, param):
        return self.fail("Invalid choice.", param=param)


def get_user_selected_account(
    account_type: Optional[Type[AccountAPI]] = None, prompt_message: Optional[str] = None
) -> AccountAPI:
    """
    Prompts the user to pick from their accounts
    and returns that account. Optionally filter the accounts
    by type.

    Use this method if you want to prompt users to select
    accounts _outside_ of CLI options. For CLI options,
    use :meth:`ape.cli.options.account_option_that_prompts_when_not_given`.
    """
    prompt = AccountAliasPromptChoice(account_type=account_type, prompt_message=prompt_message)
    return prompt.get_user_selected_account()


class AccountAliasPromptChoice(PromptChoice):
    """
    Prompts the user to select an alias from their accounts.
    Useful for adhoc scripts to lessen the need to hard-code aliases.
    """

    # noinspection PyMissingConstructor
    def __init__(self, account_type: Optional[Type[AccountAPI]] = None, prompt_message: str = None):
        # NOTE: we purposely skip the constructor of `PromptChoice`
        self._account_type = account_type
        self._prompt_message = prompt_message or "Select an account"

    # type: ignore
    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Optional[AccountAPI]:
        if value and value in accounts.aliases:
            return accounts.load(value)

        # Prompt the user if they didn't provide a value.
        alias = super().convert(value, param, ctx)
        return accounts.load(alias) if alias else None

    @property
    def choices(self) -> List[str]:
        return [
            a.alias
            for a in _get_account_by_type(self._account_type)
            if a is not None and a.alias is not None
        ]

    def get_user_selected_account(self) -> AccountAPI:
        """
        Returns the selected account.
        """
        if not self.choices:
            raise AccountsError("No accounts found.")
        elif len(self.choices) == 1:
            return accounts.load(self.choices[0])

        self.print_choices()
        return click.prompt(self._prompt_message, type=self)

    def fail_from_invalid_choice(self, param):
        return self.fail("Invalid choice. Type the number or the alias.", param=param)


class NetworkChoice(click.Choice):
    """Wraps ``click.Choice`` to provide network choice defaults for the active project."""

    def __init__(self, case_sensitive=True):
        super().__init__(list(networks.network_choices), case_sensitive)

    def get_metavar(self, param):
        return "[ecosystem-name][:[network-name][:[provider-name]]]"



from typing import Optional, Type

import click

from ape import accounts
from ape.api import AccountAPI
from ape.cli.choices import Alias
from ape.exceptions import AliasAlreadyInUseError


def _require_non_existing_alias(value):
    if value in accounts.aliases:
        raise AliasAlreadyInUseError(value)
    return value


def existing_alias_argument(account_type: Optional[Type[AccountAPI]] = None):
    return click.argument("alias", type=Alias(account_type=account_type))


def non_existing_alias_argument():
    return click.argument(
        "alias", callback=lambda ctx, param, value: _require_non_existing_alias(value)
    )



import click

from ape import networks
from ape.cli import ape_cli_context


@click.group(short_help="Manage networks")
def cli():
    """
    Command-line helper for managing networks.
    """


@cli.command(name="list", short_help="List registered networks")
@ape_cli_context()
def _list(cli_ctx):
    click.echo(networks.networks_yaml)



from pathlib import Path

import pytest

from ape.utils import extract_nested_value, get_relative_path

_TEST_DIRECTORY_PATH = Path("/This/is/a/test/")
_TEST_FILE_PATH = _TEST_DIRECTORY_PATH / "scripts" / "script.py"


def test_get_relative_path_from_project():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_DIRECTORY_PATH)
    expected = Path("scripts/script.py")
    assert actual == expected


def test_get_relative_path_given_relative_path():
    relative_script_path = Path("../deploy.py")
    with pytest.raises(ValueError) as err:
        get_relative_path(relative_script_path, _TEST_DIRECTORY_PATH)

    assert str(err.value) == "'target' must be an absolute path."

    relative_project_path = Path("../This/is/a/test")

    with pytest.raises(ValueError) as err:
        get_relative_path(_TEST_FILE_PATH, relative_project_path)

    assert str(err.value) == "'anchor' must be an absolute path."


def test_get_relative_path_same_path():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_FILE_PATH)
    assert actual == Path()


def test_get_relative_path_roots():
    root = Path("/")
    actual = get_relative_path(root, root)
    assert actual == Path()


def test_extract_nested_value():
    structure = {"foo": {"bar": {"test": "expected_value"}}}
    assert extract_nested_value(structure, "foo", "bar", "test") == "expected_value"


def test_extract_nested_value_non_dict_in_middle_returns_none():
    structure = {"foo": {"non_dict": 3, "bar": {"test": "expected_value"}}}
    assert not extract_nested_value(structure, "foo", "non_dict", "test")



from ape import plugins
from ape.api.config import ConfigItem
from ape_test.accounts import TestAccount, TestAccountContainer

from .providers import LocalNetwork


class Config(ConfigItem):
    mnemonic: str = "test test test test test test test test test test test junk"
    number_of_accounts: int = 10


@plugins.register(plugins.Config)
def config_class():
    return Config


@plugins.register(plugins.AccountPlugin)
def account_types():
    return TestAccountContainer, TestAccount


@plugins.register(plugins.ProviderPlugin)
def providers():
    yield "ethereum", "development", LocalNetwork



_DEFAULT_NETWORKS_LIST = """
ecosystems:
- name: ethereum
  isDefault: true
  networks:
  - name: mainnet
    providers:
    - name: geth
      isDefault: true
  - name: ropsten
    providers:
    - name: geth
      isDefault: true
  - name: kovan
    providers:
    - name: geth
      isDefault: true
  - name: rinkeby
    providers:
    - name: geth
      isDefault: true
  - name: goerli
    providers:
    - name: geth
      isDefault: true
  - name: development
    isDefault: true
    providers:
    - name: geth
      isDefault: true
    - name: test
"""


def test_list(ape_cli, runner):
    result = runner.invoke(ape_cli, ["networks", "list"])
    expected = _DEFAULT_NETWORKS_LIST.strip()
    assert expected in result.output



from enum import Enum, IntEnum
from pathlib import Path
from typing import Iterator, List, Optional

from dataclassy import as_dict
from hexbytes import HexBytes
from web3 import Web3

from ape.exceptions import ProviderError
from ape.logging import logger
from ape.types import TransactionSignature

from . import networks
from .base import abstractdataclass, abstractmethod
from .config import ConfigItem


class TransactionType(Enum):
    STATIC = "0x0"
    DYNAMIC = "0x2"  # EIP-1559


@abstractdataclass
class TransactionAPI:
    chain_id: int = 0
    sender: str = ""
    receiver: str = ""
    nonce: Optional[int] = None  # NOTE: `Optional` only to denote using default behavior
    value: int = 0
    gas_limit: Optional[int] = None  # NOTE: `Optional` only to denote using default behavior
    data: bytes = b""
    type: TransactionType = TransactionType.STATIC

    signature: Optional[TransactionSignature] = None

    def __post_init__(self):
        if not self.is_valid:
            raise ProviderError("Transaction is not valid.")

    @property
    def max_fee(self) -> int:
        """
        The total amount in fees willing to be spent on a transaction.
        Override this property as needed, such as for EIP-1559 differences.

        See :class:`~ape_ethereum.ecosystem.StaticFeeTransaction` and
        :class`~ape_ethereum.ecosystem.DynamicFeeTransaction` as examples.
        """
        return 0

    @max_fee.setter
    def max_fee(self, value):
        raise NotImplementedError("Max fee is not settable by default.")

    @property
    def total_transfer_value(self) -> int:
        """
        The total amount of WEI that a transaction could use.
        Useful for determining if an account balance can afford
        to submit the transaction.
        """
        return self.value + self.max_fee

    @property
    @abstractmethod
    def is_valid(self):
        ...

    @abstractmethod
    def encode(self) -> bytes:
        """
        Take this object and produce a hash to sign to submit a transaction
        """

    def as_dict(self) -> dict:
        return as_dict(self)

    def __repr__(self) -> str:
        data = as_dict(self)  # NOTE: `as_dict` could be overridden
        params = ", ".join(f"{k}={v}" for k, v in data.items())
        return f"<{self.__class__.__name__} {params}>"

    def __str__(self) -> str:
        data = as_dict(self)  # NOTE: `as_dict` could be overridden
        if len(data["data"]) > 9:
            data["data"] = (
                "0x" + bytes(data["data"][:3]).hex() + "..." + bytes(data["data"][-3:]).hex()
            )
        else:
            data["data"] = "0x" + bytes(data["data"]).hex()
        params = "\n  ".join(f"{k}: {v}" for k, v in data.items())
        return f"{self.__class__.__name__}:\n  {params}"


class TransactionStatusEnum(IntEnum):
    FAILING = 0
    NO_ERROR = 1


@abstractdataclass
class ReceiptAPI:
    txn_hash: str
    status: TransactionStatusEnum
    block_number: int
    gas_used: int
    gas_price: int
    logs: List[dict] = []
    contract_address: Optional[str] = None

    def __post_init__(self):
        txn_hash = self.txn_hash.hex() if isinstance(self.txn_hash, HexBytes) else self.txn_hash
        logger.info(f"Submitted {txn_hash} (gas_used={self.gas_used})")

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.txn_hash}>"

    def raise_for_status(self, txn: TransactionAPI):
        """
        Handle provider-specific errors regarding a non-successful
        :class:`~api.providers.TransactionStatusEnum`.
        """

    def ran_out_of_gas(self, gas_limit: int) -> bool:
        """
        Returns ``True`` when the transaction failed and used the
        same amount of gas as the given ``gas_limit``.
        """
        return self.status == TransactionStatusEnum.FAILING and self.gas_used == gas_limit

    @classmethod
    @abstractmethod
    def decode(cls, data: dict) -> "ReceiptAPI":
        ...


@abstractdataclass
class ProviderAPI:
    """
    A Provider must work with a particular Network in a particular Ecosystem
    """

    name: str  # Plugin name
    network: networks.NetworkAPI
    config: ConfigItem
    provider_settings: dict
    data_folder: Path
    request_header: str

    @abstractmethod
    def connect(self):
        ...

    @abstractmethod
    def disconnect(self):
        ...

    @abstractmethod
    def update_settings(self, new_settings: dict):
        ...

    @property
    @abstractmethod
    def chain_id(self) -> int:
        ...

    @abstractmethod
    def get_balance(self, address: str) -> int:
        ...

    @abstractmethod
    def get_code(self, address: str) -> bytes:
        ...

    @abstractmethod
    def get_nonce(self, address: str) -> int:
        ...

    @abstractmethod
    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        ...

    @property
    @abstractmethod
    def gas_price(self) -> int:
        ...

    @property
    @abstractmethod
    def priority_fee(self) -> int:
        raise NotImplementedError("priority_fee is not implemented by this provider")

    @property
    @abstractmethod
    def base_fee(self) -> int:
        raise NotImplementedError("base_fee is not implemented by this provider")

    @abstractmethod
    def send_call(self, txn: TransactionAPI) -> bytes:  # Return value of function
        ...

    @abstractmethod
    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        ...

    @abstractmethod
    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        ...

    @abstractmethod
    def get_events(self, **filter_params) -> Iterator[dict]:
        ...


class TestProviderAPI(ProviderAPI):
    """
    An API for providers that have development functionality, such as snapshotting.
    """

    @abstractmethod
    def snapshot(self) -> str:
        ...

    @abstractmethod
    def revert(self, snapshot_id: str):
        ...


class Web3Provider(ProviderAPI):
    """
    A base provider that is web3 based.
    """

    _web3: Web3 = None  # type: ignore

    def update_settings(self, new_settings: dict):
        """
        Update the provider settings and re-connect.
        """
        self.disconnect()
        self.provider_settings.update(new_settings)
        self.connect()

    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        """
        Generates and returns an estimate of how much gas is necessary
        to allow the transaction to complete.
        The transaction will not be added to the blockchain.
        """
        txn_dict = txn.as_dict()
        return self._web3.eth.estimate_gas(txn_dict)  # type: ignore

    @property
    def chain_id(self) -> int:
        """
        Returns the currently configured chain ID,
        a value used in replay-protected transaction signing as introduced by EIP-155.
        """
        return self._web3.eth.chain_id

    @property
    def gas_price(self) -> int:
        """
        Returns the current price per gas in wei.
        """
        return self._web3.eth.generate_gas_price()  # type: ignore

    @property
    def priority_fee(self) -> int:
        """
        Returns the current max priority fee per gas in wei.
        """
        return self._web3.eth.max_priority_fee

    @property
    def base_fee(self) -> int:
        block = self._web3.eth.get_block("latest")
        return block.baseFeePerGas  # type: ignore

    def get_nonce(self, address: str) -> int:
        """
        Returns the number of transactions sent from an address.
        """
        return self._web3.eth.get_transaction_count(address)  # type: ignore

    def get_balance(self, address: str) -> int:
        """
        Returns the balance of the account of a given address.
        """
        return self._web3.eth.get_balance(address)  # type: ignore

    def get_code(self, address: str) -> bytes:
        """
        Returns code at a given address.
        """
        return self._web3.eth.get_code(address)  # type: ignore

    def send_call(self, txn: TransactionAPI) -> bytes:
        """
        Executes a new message call immediately without creating a
        transaction on the block chain.
        """
        return self._web3.eth.call(txn.as_dict())

    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        """
        Returns the information about a transaction requested by transaction hash.
        """
        # TODO: Work on API that let's you work with ReceiptAPI and re-send transactions
        receipt = self._web3.eth.wait_for_transaction_receipt(txn_hash)  # type: ignore
        txn = self._web3.eth.get_transaction(txn_hash)  # type: ignore
        return self.network.ecosystem.receipt_class.decode({**txn, **receipt})

    def get_events(self, **filter_params) -> Iterator[dict]:
        """
        Returns an array of all logs matching a given set of filter parameters.
        """
        return iter(self._web3.eth.get_logs(filter_params))  # type: ignore

    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        txn_hash = self._web3.eth.send_raw_transaction(txn.encode())
        receipt = self.get_transaction(txn_hash.hex())
        return receipt



from typing import List, Optional, Union

import click

from ape import networks, project
from ape.cli.choices import AccountAliasPromptChoice, NetworkChoice
from ape.cli.utils import Abort
from ape.exceptions import ContractError
from ape.logging import LogLevel, logger
from ape.types import ContractType


class ApeCliContextObject:
    """A class that can be auto-imported into a plugin ``click.command()``
    via ``@ape_cli_context()``. It can help do common CLI tasks such as log
    messages to the user or abort execution."""

    def __init__(self):
        self.logger = logger

    @staticmethod
    def abort(msg: str, base_error: Exception = None):
        if base_error:
            logger.error(msg)
            raise Abort(msg) from base_error
        raise Abort(msg)


def verbosity_option(cli_logger):
    """A decorator that adds a `--verbosity, -v` option to the decorated
    command.
    """

    level_names = [lvl.name for lvl in LogLevel]
    names_str = f"{', '.join(level_names[:-1])}, or {level_names[-1]}"

    def decorator(f):
        def _set_level(ctx, param, value):
            log_level = getattr(LogLevel, value.upper(), None)
            if log_level is None:
                raise click.BadParameter(f"Must be one of {names_str}, not {value}.")

            cli_logger.set_level(log_level.name)

        return click.option(
            "--verbosity",
            "-v",
            callback=_set_level,
            default=LogLevel.INFO.name,
            metavar="LVL",
            expose_value=False,
            help=f"One of {names_str}",
            is_eager=True,
        )(f)

    return decorator


def ape_cli_context():
    def decorator(f):
        f = verbosity_option(logger)(f)
        f = click.make_pass_decorator(ApeCliContextObject, ensure=True)(f)
        return f

    return decorator


def network_option():
    return click.option(
        "--network",
        type=NetworkChoice(case_sensitive=False),
        default=networks.default_ecosystem.name,
        help="Override the default network and provider. (see ``ape networks list`` for options)",
        show_default=True,
        show_choices=False,
    )


def skip_confirmation_option(help=""):
    return click.option(
        "-y",
        "--yes",
        "skip_confirmation",
        default=False,
        is_flag=True,
        help=help,
    )


def _account_callback(ctx, param, value):
    if param and not value:
        return param.type.get_user_selected_account()
    return value


def account_option_that_prompts_when_not_given():
    """
    Accepts either the account alias or the account number.
    If not given anything, it will prompt the user to select an account.
    """
    return click.option(
        "--account",
        type=AccountAliasPromptChoice(),
        callback=_account_callback,
    )


def _load_contracts(ctx, param, value) -> Optional[Union[ContractType, List[ContractType]]]:
    if not value:
        return None

    if len(project.contracts) == 0:
        raise ContractError("Project has no contracts.")

    # If the user passed in `multiple=True`, then `value` is a list,
    # and therefore we should also return a list.
    is_multiple = isinstance(value, (tuple, list))

    def create_contract(contract_name: str) -> ContractType:
        if contract_name not in project.contracts:
            raise ContractError(f"No contract named '{value}'")

        return project.contracts[contract_name]

    return [create_contract(c) for c in value] if is_multiple else create_contract(value)


def contract_option(help=None, required=False, multiple=False):
    """
    Contract(s) from the current project.
    If you pass ``multiple=True``, you will get a list of contract types from the callback.
    """
    help = help or "The name of a contract in the current project"
    return click.option(
        "--contract", help=help, required=required, callback=_load_contracts, multiple=multiple
    )



import pytest

from ape.api import (
    AccountContainerAPI,
    EcosystemAPI,
    NetworkAPI,
    ProviderAPI,
    ReceiptAPI,
    TransactionAPI,
    TransactionStatusEnum,
)
from ape.api.config import ConfigItem
from ape.exceptions import ContractLogicError

TEST_ADDRESS = "0x0A78AAAAA2122100000b9046f0A085AB2E111113"


@pytest.fixture
def mock_account_container_api(mocker):
    return mocker.MagicMock(spec=AccountContainerAPI)


@pytest.fixture
def mock_provider_api(mocker, mock_network_api):
    mock = mocker.MagicMock(spec=ProviderAPI)
    mock.network = mock_network_api
    return mock


class _ContractLogicError(ContractLogicError):
    pass


@pytest.fixture
def mock_network_api(mocker):
    mock = mocker.MagicMock(spec=NetworkAPI)
    mock_ecosystem = mocker.MagicMock(spec=EcosystemAPI)
    mock_ecosystem.virtual_machine_error_class = _ContractLogicError
    mock.ecosystem = mock_ecosystem
    return mock


@pytest.fixture
def mock_failing_transaction_receipt(mocker):
    mock = mocker.MagicMock(spec=ReceiptAPI)
    mock.status = TransactionStatusEnum.FAILING
    mock.gas_used = 0
    return mock


@pytest.fixture
def mock_web3(mocker):
    return mocker.MagicMock()


@pytest.fixture
def mock_config_item(mocker):
    return mocker.MagicMock(spec=ConfigItem)


@pytest.fixture
def mock_transaction(mocker):
    return mocker.MagicMock(spec=TransactionAPI)



from typing import Iterator, List, Optional

from eth_account import Account as EthAccount  # type: ignore
from eth_account.messages import SignableMessage

from ape.api import TestAccountAPI, TestAccountContainerAPI, TransactionAPI
from ape.convert import to_address
from ape.types import AddressType, MessageSignature, TransactionSignature
from ape.utils import GeneratedDevAccount, cached_property, generate_dev_accounts


class TestAccountContainer(TestAccountContainerAPI):
    @property
    def config(self):
        return self.config_manager.get_config("test")

    @cached_property
    def _dev_accounts(self) -> List[GeneratedDevAccount]:
        mnemonic = self.config["mnemonic"]
        return generate_dev_accounts(mnemonic)

    @property
    def aliases(self) -> Iterator[str]:
        for index in range(0, len(self)):
            yield f"dev_{index}"

    def __len__(self) -> int:
        return len(self._dev_accounts)

    def __iter__(self) -> Iterator[TestAccountAPI]:
        for index in range(0, len(self)):
            account = self._dev_accounts[index]
            yield TestAccount(
                self, _index=index, _address=account.address, _private_key=account.private_key
            )  # type: ignore


class TestAccount(TestAccountAPI):
    _index: int
    _address: str
    _private_key: str

    @property
    def alias(self) -> str:
        return f"dev_{self._index}"

    @property
    def address(self) -> AddressType:
        return to_address(self._address)

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        signed_msg = EthAccount.sign_message(msg, self._private_key)
        return MessageSignature(v=signed_msg.v, r=signed_msg.r, s=signed_msg.s)  # type: ignore

    def sign_transaction(self, txn: TransactionAPI) -> Optional[TransactionSignature]:
        signed_txn = EthAccount.sign_transaction(txn.as_dict(), self._private_key)
        return TransactionSignature(v=signed_txn.v, r=signed_txn.r, s=signed_txn.s)  # type: ignore



from pathlib import Path

import pytest
from web3.exceptions import ContractLogicError as Web3ContractLogicError

from ape.api import ReceiptAPI, TransactionStatusEnum
from ape.exceptions import ContractLogicError, TransactionError
from ape_geth import GethProvider

_TEST_REVERT_REASON = "TEST REVERT REASON."


def _create_mock_receipt(status=TransactionStatusEnum.NO_ERROR, gas_used=0):
    class MockReceipt(ReceiptAPI):
        @classmethod
        def decode(cls, data: dict) -> "ReceiptAPI":
            return MockReceipt(
                txn_hash="test-hash",  # type: ignore
                status=status,  # type: ignore
                gas_used=gas_used,  # type: ignore
                gas_price="60000000000",  # type: ignore
                block_number=0,  # type: ignore
            )

    return MockReceipt


class TestEthereumProvider:
    def test_send_when_web3_error_raises_transaction_error(
        self, mock_web3, mock_network_api, mock_config_item, mock_transaction
    ):
        provider = GethProvider(
            name="test",
            network=mock_network_api,
            config=mock_config_item,
            provider_settings={},
            data_folder=Path("."),
            request_header="",
        )
        provider._web3 = mock_web3

        mock_network_api.ecosystem.receipt_class = _create_mock_receipt()
        web3_error_data = {
            "code": -32000,
            "message": "Test Error Message",
        }
        mock_web3.eth.send_raw_transaction.side_effect = ValueError(web3_error_data)
        with pytest.raises(TransactionError) as err:
            provider.send_transaction(mock_transaction)

        assert web3_error_data["message"] in str(err.value)

    def test_send_transaction_reverts_from_contract_logic(
        self, mock_web3, mock_network_api, mock_config_item, mock_transaction
    ):
        provider = GethProvider(
            name="test",
            network=mock_network_api,
            config=mock_config_item,
            provider_settings={},
            data_folder=Path("."),
            request_header="",
        )
        provider._web3 = mock_web3
        mock_network_api.ecosystem.receipt_class = _create_mock_receipt()
        test_err = Web3ContractLogicError(f"execution reverted: {_TEST_REVERT_REASON}")
        mock_web3.eth.send_raw_transaction.side_effect = test_err

        with pytest.raises(ContractLogicError) as err:
            provider.send_transaction(mock_transaction)

        assert str(err.value) == _TEST_REVERT_REASON



from .utils import skip_projects, skip_projects_except


@skip_projects(["unregistered-contracts", "one-interface"])
def test_compile_missing_contracts_dir(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    assert "WARNING" in result.output
    assert "No 'contracts/' directory detected" in result.output


@skip_projects_except(["unregistered-contracts"])
def test_missing_extensions(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    assert "WARNING: No compilers detected for the " "following extensions:" in result.output
    assert ".test" in result.output
    assert ".foobar" in result.output

    result = runner.invoke(ape_cli, ["compile", "contracts/Contract.test"])
    assert result.exit_code == 0
    assert (
        "WARNING: No compilers detected for the " "following extensions: .test"
    ) in result.output


@skip_projects(["empty-config", "no-config", "script", "unregistered-contracts"])
def test_compile(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    # First time it compiles, it compiles fully
    for file in project.path.glob("contracts/**/*"):
        assert file.stem in result.output

    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    # First time it compiles, it caches
    for file in project.path.glob("contracts/**/*"):
        assert file.stem not in result.output


@skip_projects_except([])
def test_compile_contracts(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["compile", "--size"])
    assert result.exit_code == 0
    # Still caches but displays bytecode size
    for file in project.path.glob("contracts/**/*"):
        assert file.stem in result.output



from typing import List

import pytest

from ape.api import ProviderAPI, TestAccountAPI
from ape.exceptions import ProviderError
from ape.managers.accounts import AccountManager
from ape.managers.networks import NetworkManager
from ape.managers.project import ProjectManager


class PytestApeFixtures:
    def __init__(self, accounts: AccountManager, networks: NetworkManager, project: ProjectManager):
        self._accounts = accounts
        self._networks = networks
        self._project = project

    @pytest.fixture
    def accounts(self, provider) -> List[TestAccountAPI]:
        return self._accounts.test_accounts

    @pytest.fixture
    def provider(self) -> ProviderAPI:
        active_provider = self._networks.active_provider

        if active_provider is None:
            raise ProviderError("Provider is not set.")

        return active_provider

    @pytest.fixture(scope="session")
    def project(self) -> ProjectManager:
        return self._project



from typing import Optional


class ApeException(Exception):
    """
    An exception raised by ape.
    """

    def __init__(self, message):
        if not message.endswith("."):
            message = f"{message}."
        super().__init__(message)


class AccountsError(ApeException):
    """
    Raised when a problem occurs when using accounts.
    """


class AliasAlreadyInUseError(AccountsError):
    """
    Raised when attempting to add an account using an alias
    that already maps to another account.
    """

    def __init__(self, alias: str):
        self.alias = alias
        super().__init__(f"Account with alias '{alias}' already in use.")


class SignatureError(AccountsError):
    """
    Raised when there are issues with signing.
    """


class ContractError(ApeException):
    """
    Raised when issues occur when interacting with a contract
    (calls or transactions).
    """


class ArgumentsLengthError(ContractError):
    """
    Raised when calling a contract method with the wrong number of arguments.
    """

    def __init__(self):
        message = "The number of the given arguments do not match what is defined in the ABI."
        super().__init__(message)


class DecodingError(ContractError):
    """
    Raised when issues occur while decoding data from
    a contract call or transaction.
    """

    def __init__(self):
        super().__init__("Output corrupted.")


class TransactionError(ContractError):
    """
    Raised when issues occur related to transactions.
    """

    def __init__(
        self,
        base_err: Optional[Exception] = None,
        message: Optional[str] = None,
        code: Optional[int] = None,
    ):
        self.base_err = base_err
        if not message:
            message = str(base_err) if base_err else "Tranaction Failed"

        self.message = message
        self.code = code

        ex_message = f"({code}) {message}" if code else message
        super().__init__(ex_message)


class VirtualMachineError(TransactionError):
    """
    Raised when a transaction error occurs in a virtual machine.
    """


class ContractLogicError(VirtualMachineError):
    """
    Raised when there is a contract-defined revert,
    such as from an assert/require statement.
    """

    def __init__(self, revert_message: str):
        super().__init__(message=revert_message)

    @property
    def revert_message(self):
        return self.message

    @classmethod
    def from_error(cls, err: Exception):
        """
        Creates this class from the error message of the given
        error.

        This should be overridden whenever possible to handle
        provider-specific use-cases for raising this error.
        """
        return cls(str(err))


class OutOfGasError(TransactionError):
    """
    Raised when detecting a transaction failed because it ran
    out of gas.
    """

    def __init__(self, code: Optional[int] = None):
        super().__init__(message="The transaction ran out of gas.", code=code)


class ContractDeployError(TransactionError):
    """
    Raised when a problem occurs when deploying a contract.
    """


class NetworkError(ApeException):
    """
    Raised when a problem occurs when using blockchain networks.
    """


class NetworkNotFoundError(NetworkError):
    """
    Raised when the network with the given name was not found.
    """

    def __init__(self, network: str):
        self.network = network
        message = f"No network named '{network}'."
        super().__init__(message)


class CompilerError(ApeException):
    """
    Raised when unable to compile.
    """


class ProjectError(ApeException):
    """
    Raised when problems occur in a project.
    """


class ConversionError(ApeException):
    """
    Raised when unable to convert a value.
    """


class ProviderError(ApeException):
    """
    Raised when a problem occurs when using providers.
    """


class ConfigError(ApeException):
    """
    Raised when a problem occurs from the configuration file.
    """


class AddressError(ApeException):
    """
    Raised when a problem occurs regarding an address.
    """



from .utils import skip_projects

BAD_COMMAND = "not-a-name"


@skip_projects(["script"])
def test_run_no_scripts_dir(ape_cli, runner, project):
    result = runner.invoke_using_test_network(ape_cli, ["run", BAD_COMMAND])
    assert result.exit_code == 1
    assert "No 'scripts/' directory detected to run script" in result.output


@skip_projects(["empty-config", "no-config", "one-interface", "unregistered-contracts"])
def test_run(ape_cli, runner, project):
    result = runner.invoke_using_test_network(ape_cli, ["run"])
    assert result.exit_code == 1, result.output
    assert "Must provide at least one script name or path" in result.output

    result = runner.invoke_using_test_network(ape_cli, ["run", BAD_COMMAND])
    assert result.exit_code == 1
    assert f"No script named '{BAD_COMMAND}' detected in scripts folder" in result.output

    for script_file in (project.path / "scripts").glob("*.py"):
        if script_file.stem == "__init__":
            continue

        result = runner.invoke_using_test_network(ape_cli, ["run", script_file.stem])
        assert result.exit_code == 0, result.output



from .accounts import AccountAPI, AccountContainerAPI, TestAccountAPI, TestAccountContainerAPI
from .address import Address, AddressAPI
from .contracts import ContractInstance, ContractLog
from .convert import ConverterAPI
from .explorers import ExplorerAPI
from .networks import EcosystemAPI, NetworkAPI, ProviderContextManager, create_network_type
from .providers import (
    ProviderAPI,
    ReceiptAPI,
    TestProviderAPI,
    TransactionAPI,
    TransactionStatusEnum,
    TransactionType,
    Web3Provider,
)

__all__ = [
    "AccountAPI",
    "AccountContainerAPI",
    "Address",
    "AddressAPI",
    "ContractInstance",
    "ContractLog",
    "ConverterAPI",
    "create_network_type",
    "EcosystemAPI",
    "ExplorerAPI",
    "ProviderAPI",
    "ProviderContextManager",
    "NetworkAPI",
    "ReceiptAPI",
    "TestAccountAPI",
    "TestAccountContainerAPI",
    "TestProviderAPI",
    "TransactionAPI",
    "TransactionStatusEnum",
    "TransactionType",
    "Web3Provider",
]



import json
from pathlib import Path
from typing import Dict, List, Optional

import requests
from dataclassy import dataclass

from ape.exceptions import ProjectError
from ape.types import Checksum, Compiler, ContractType, PackageManifest, Source
from ape.utils import compute_checksum

from .compilers import CompilerManager
from .config import ConfigManager


@dataclass
class ProjectManager:
    path: Path
    config: ConfigManager
    compilers: CompilerManager

    dependencies: Dict[str, PackageManifest] = dict()

    def __post_init__(self):
        if isinstance(self.path, str):
            self.path = Path(self.path)

        self.dependencies = {
            manifest.name: manifest
            for manifest in map(self._extract_manifest, self.config.dependencies)
        }

    def _extract_manifest(self, manifest_uri: str) -> PackageManifest:
        manifest_dict = requests.get(manifest_uri).json()
        # TODO: Handle non-manifest URLs e.g. Ape/Brownie projects, Hardhat/Truffle projects, etc.
        if "name" not in manifest_dict:
            raise ProjectError("Dependencies must have a name.")

        return PackageManifest.from_dict(manifest_dict)

    def __str__(self) -> str:
        return f'Project("{self.path}")'

    @property
    def _cache_folder(self) -> Path:
        folder = self.path / ".build"
        # NOTE: If we use the cache folder, we expect it to exist
        folder.mkdir(exist_ok=True)
        return folder

    @property
    def manifest_cachefile(self) -> Path:
        file_name = self.config.name or "__local__"
        return self._cache_folder / (file_name + ".json")

    @property
    def cached_manifest(self) -> Optional[PackageManifest]:
        manifest_file = self.manifest_cachefile
        if manifest_file.exists():
            manifest_json = json.loads(manifest_file.read_text())
            if "manifest" not in manifest_json:
                raise ProjectError("Corrupted manifest.")

            return PackageManifest.from_dict(manifest_json)

        else:
            return None

    # NOTE: Using these paths should handle the case when the folder doesn't exist
    @property
    def contracts_folder(self) -> Path:
        return self.path / "contracts"

    @property
    def sources(self) -> List[Path]:
        """All the source files in the project.
        Excludes files with extensions that don't have a registered compiler.
        """
        files: List[Path] = []
        for extension in self.compilers.registered_compilers:
            files.extend(self.contracts_folder.rglob("*" + extension))

        return files

    @property
    def sources_missing(self) -> bool:
        return not self.contracts_folder.exists() or not self.contracts_folder.iterdir()

    @property
    def extensions_with_missing_compilers(self) -> List[str]:
        """All file extensions in the `contracts/` directory (recursively)
        that do not correspond to a registered compiler."""
        extensions = []

        def _append_extensions_in_dir(directory: Path):
            for file in directory.iterdir():
                if file.is_dir():
                    _append_extensions_in_dir(file)
                elif (
                    file.suffix
                    and file.suffix not in extensions
                    and file.suffix not in self.compilers.registered_compilers
                ):
                    extensions.append(file.suffix)

        _append_extensions_in_dir(self.contracts_folder)
        return extensions

    def load_contracts(self, use_cache: bool = True) -> Dict[str, ContractType]:
        # Load a cached or clean manifest (to use for caching)
        manifest = use_cache and self.cached_manifest or PackageManifest()
        cached_sources = manifest.sources or {}
        cached_contract_types = manifest.contractTypes or {}

        # If a file is deleted from ``self.sources`` but is in
        # ``cached_sources``, remove its corresponding ``contract_types`` by
        # using ``ContractType.sourceId`` and ``ContractType.sourcePath``
        deleted_sources = cached_sources.keys() - set(map(str, self.sources))
        contract_types = {}
        for name, ct in cached_contract_types.items():
            if ct.sourceId in deleted_sources:
                pass  # this contract's source code file was deleted
            else:
                contract_types[name] = ct

        def file_needs_compiling(source: Path) -> bool:
            path = str(source)
            # New file added?
            if path not in cached_sources:
                return True

            # Recalculate checksum if it doesn't exist yet
            cached = cached_sources[path]
            cached.compute_checksum(algorithm="md5")

            assert cached.checksum  # to tell mypy this can't be None

            # File contents changed in source code folder?
            checksum = compute_checksum(
                source.read_bytes(),
                algorithm=cached.checksum.algorithm,
            )
            return checksum != cached.checksum.hash

        # NOTE: filter by checksum, etc., and compile what's needed
        #       to bring our cached manifest up-to-date
        needs_compiling = filter(file_needs_compiling, self.sources)
        contract_types.update(self.compilers.compile(list(needs_compiling)))

        # Update cached contract types & source code entries in cached manifest
        manifest.contractTypes = contract_types
        cached_sources = {
            str(source): Source(  # type: ignore
                checksum=Checksum(  # type: ignore
                    algorithm="md5", hash=compute_checksum(source.read_bytes())
                ),
                urls=[],
            )
            for source in self.sources
        }
        manifest.sources = cached_sources

        # NOTE: Cache the updated manifest to disk (so ``self.cached_manifest`` reads next time)
        self.manifest_cachefile.write_text(json.dumps(manifest.to_dict()))

        return contract_types

    @property
    def contracts(self) -> Dict[str, ContractType]:
        return self.load_contracts()

    def __getattr__(self, attr_name: str):
        contracts = self.load_contracts()
        if attr_name in contracts:
            return contracts[attr_name]
        elif attr_name in self.dependencies:
            return self.dependencies[attr_name]
        else:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'.")

    @property
    def interfaces_folder(self) -> Path:
        return self.path / "interfaces"

    @property
    def scripts_folder(self) -> Path:
        return self.path / "scripts"

    @property
    def tests_folder(self) -> Path:
        return self.path / "tests"

    # TODO: Make this work for generating and caching the manifest file
    @property
    def compiler_data(self) -> List[Compiler]:
        compilers = []

        for extension, compiler in self.compilers.registered_compilers.items():
            for version in compiler.get_versions(
                [p for p in self.sources if p.suffix == extension]
            ):
                compilers.append(Compiler(compiler.name, version))  # type: ignore

        return compilers

    # @property
    # def meta(self) -> PackageMeta:
    #     return PackageMeta(**self.config.get_config("ethpm").serialize())

    # def publish_manifest(self):
    #     manifest = self.manifest.to_dict()  # noqa: F841
    #     if not manifest["name"]:
    #         raise ProjectError("Need name to release manifest")
    #     if not manifest["version"]:
    #         raise ProjectError("Need version to release manifest")
    #     # TODO: Clean up manifest and minify it
    #     # TODO: Publish sources to IPFS and replace with CIDs
    #     # TODO: Publish to IPFS



import collections
import json
import os
from copy import deepcopy
from functools import lru_cache
from hashlib import md5
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional

import yaml
from eth_account import Account
from eth_account.hdaccount import HDPath, seed_from_mnemonic
from hexbytes import HexBytes
from importlib_metadata import PackageNotFoundError, packages_distributions, version

from ape.logging import logger

try:
    from functools import cached_property  # type: ignore
except ImportError:
    from backports.cached_property import cached_property  # type: ignore

try:
    from functools import singledispatchmethod  # type: ignore
except ImportError:
    from singledispatchmethod import singledispatchmethod  # type: ignore


@lru_cache(maxsize=None)
def get_distributions():
    return packages_distributions()


def is_relative_to(path: Path, target: Path) -> bool:
    if hasattr(path, "is_relative_to"):
        # NOTE: Only available ``>=3.9``
        return target.is_relative_to(path)  # type: ignore

    else:
        try:
            return target.relative_to(path) is not None
        except ValueError:
            return False


def get_relative_path(target: Path, anchor: Path) -> Path:
    """
    Compute the relative path of ``target`` relative to ``anchor``,
    which may or may not share a common ancestor.
    NOTE: Both paths must be absolute
    """
    if not target.is_absolute():
        raise ValueError("'target' must be an absolute path.")
    if not anchor.is_absolute():
        raise ValueError("'anchor' must be an absolute path.")

    anchor_copy = Path(str(anchor))
    levels_deep = 0
    while not is_relative_to(anchor_copy, target):
        levels_deep += 1
        anchor_copy = anchor_copy.parent

    return Path("/".join(".." for _ in range(levels_deep))).joinpath(
        str(target.relative_to(anchor_copy))
    )


def get_package_version(obj: Any) -> str:
    # If value is already cached/static
    if hasattr(obj, "__version__"):
        return obj.__version__

    # NOTE: In case where don't pass a module name
    if not isinstance(obj, str):
        obj = obj.__name__

    # Reduce module string to base package
    # NOTE: Assumed that string input is module name e.g. ``__name__``
    pkg_name = obj.split(".")[0]

    # NOTE: In case the distribution and package name differ
    dists = get_distributions()
    if pkg_name in dists:
        # NOTE: Shouldn't really be more than 1, but never know
        if len(dists[pkg_name]) != 1:
            logger.warning(f"duplicate pkg_name '{pkg_name}'")
        pkg_name = dists[pkg_name][0]

    try:
        return version(pkg_name)

    except PackageNotFoundError:
        # NOTE: Must handle empty string result here
        return ""


def deep_merge(dict1, dict2):
    """Return a new dictionary by merging two dictionaries recursively."""

    result = deepcopy(dict1)

    for key, value in dict2.items():
        if isinstance(value, collections.Mapping):
            result[key] = deep_merge(result.get(key, {}), value)
        else:
            result[key] = deepcopy(dict2[key])

    return result


def expand_environment_variables(contents: str) -> str:
    return os.path.expandvars(contents)


def load_config(path: Path, expand_envars=True, must_exist=False) -> Dict:
    if path.exists():
        contents = path.read_text()
        if expand_envars:
            contents = expand_environment_variables(contents)

        if path.suffix in (".json",):
            config = json.loads(contents)
        elif path.suffix in (".yml", ".yaml"):
            config = yaml.safe_load(contents)
        else:
            raise TypeError(f"Cannot parse '{path.suffix}' files!")

        return config or {}

    elif must_exist:
        raise OSError(f"{path} does not exist!")

    else:
        return {}


def compute_checksum(source: bytes, algorithm: str = "md5") -> str:
    if algorithm == "md5":
        hasher = md5
    else:
        raise ValueError(f"Unknown algorithm `{algorithm}`.")

    return hasher(source).hexdigest()


GeneratedDevAccount = collections.namedtuple("GeneratedDevAccount", ("address", "private_key"))


def generate_dev_accounts(
    mnemonic: str,
    number_of_accounts: int = 10,
    hd_path_format="m/44'/60'/0'/{}",
) -> List[GeneratedDevAccount]:
    """
    Creates accounts from the configured test mnemonic.
    Use these accounts (or the mnemonic) in chain-genesis
    for testing providers.
    """
    seed = seed_from_mnemonic(mnemonic, "")
    accounts = []

    for i in range(0, number_of_accounts):
        hd_path = HDPath(hd_path_format.format(i))
        private_key = HexBytes(hd_path.derive(seed)).hex()
        address = Account.from_key(private_key).address
        accounts.append(GeneratedDevAccount(address, private_key))

    return accounts


def gas_estimation_error_message(tx_error: Exception) -> str:
    """
    Use this method in ``ProviderAPI`` implementations when error handling
    transaction errors. This is to have a consistent experience across providers.
    """
    return (
        f"Gas estimation failed: '{tx_error}'. This transaction will likely revert. "
        "If you wish to broadcast, you must set the gas limit manually."
    )


def extract_nested_value(root: Mapping, *args: str) -> Optional[Dict]:
    """
    Dig through a nested ``Dict`` gives the keys to use in order as arguments.
    Returns the final value if it exists else `None` if the tree ends at any point.
    """
    current_value: Any = root
    for arg in args:
        if not hasattr(current_value, "get"):
            return None

        current_value = current_value.get(arg)

    return current_value


__all__ = [
    "cached_property",
    "deep_merge",
    "expand_environment_variables",
    "extract_nested_value",
    "get_relative_path",
    "gas_estimation_error_message",
    "GeneratedDevAccount",
    "generate_dev_accounts",
    "load_config",
    "singledispatchmethod",
]
